import logging 
from pprint import pprint 

import pandas as pd 
import pyodbc
from .exceptions import (
    QuickBooksError, QBCreateMethod, QBMissingTable, QBConnectionError
)


class QuickBooks: 
    pyodbc.pooling = False

    def __init__(self, DSN=None, remote_dsn=None, ip='', port=4500): 
        """Create a local or remote connection"""
        if DSN:
            self.connection_string = ("DSN=" + DSN)
        else: 
            self.connection_string = ';'.join([
                "Driver={QRemote for QuickBooks}",
                f"IPAddress={ip}",
                f"Port={port}",
                f"RemoteDSN={remote_dsn};"
            ])
        self.cnxn = None
        self.cursor = None

    def connect(self):
        """Initialize the connection and set the curor to be used in all calls.""" 
        try:
            if self.cnxn is None:
                logging.info('Connecting to: ', self.connection_string)
                self.cnxn = pyodbc.connect(self.connection_string, ansi=True, autocommit=True)
                self.cnxn.setencoding(encoding='utf-8')
                self.cursor = self.cnxn.cursor()
            else: 
                """Reconnect"""
                self.cursor = self.cnxn.cursor()

        except Exception as e:
            raise QBConnectionError('Could not connect to: ', self.connection_string)


    def query(self, query):
        """Raw sql returns response as a pandas df""" 
        logging.info("Preparing query for\n", query)
        try: 
            return pd.read_sql(query, self.cnxn)
        except Exception as e: 
            raise QBMissingTable('Not a table...')

    def sql(self, sql, params=[]):
        """Execute some raw sql. Params passes as a list."""
        return self.cursor.execute(sql, params)

    def last_insert(self, table): 
        """Fetch last insert"""
        return self.cursor.execute(f'sp_lastinsertid {table}').fetchval()

    def columns(self, table): 
        return self.query(f'sp_columns {table}')[
            ['COLUMNNAME', 'TYPENAME', 'NULLABLE',
       'REMARKS', 'DEFAULT', 'DATATYPE', 'DATETIME_SUBTYPE',
       'ORDINAL_POSITION', 'IS_NULLABLE', 'QUERYABLE', 'UPDATEABLE',
       'INSERTABLE', 'REQUIRED_ON_INSERT', 'FORMAT', 'RELATES_TO',
       'JUMPIN_TYPE', 'CUSTOM_FIELD_NAME']
        ]


    def primary_keys(self, table): 
        return self.cursor.execute(f'sp_primarykeys {table}').fetchall()
    
    
    def optimize_full_sync(self, table=None):
        if table is None: 
            table = 'all'
        self.cursor.execute('sp_optimizefullsync {table}')

    def tables(self): 
        cc = [list((c[2], c[4])) for c in self.cursor.execute('sp_tables').fetchall()]
        print(*cc, sep='\n')
    
    
    def file_name(self): 
        return self.cursor.execute('sp_qbfilename').fetchval()

    def delete(self, table, transaction_id): 
        return self.cursor.execute(f'sp_void {table} where TxnID = ?', transaction_id).fetchval()

    def close(self):
        """Close the connection to the current QuickBooks instance.
        Modify the DSN to prevent the current session from being closed.
        Will still close the file when opening a closed file.""" 
        self.cnxn.close()

    def __close__(self): 
        """Not sure which methods need to be implemented yet."""
        self.cnxn.close()


    # def create(self, obj): 
    #     """Takes a QuickBooks object: customer, account, bill, jouranl entry, 
    #     with the _create() method implemented. 
    #     """
    #     try:
    #         cx = self.cnxn()
    #     # Checks the _create() method is implemented or raises an exception
    #         return obj._create(cx)
    #     except AttributeError:
    #         raise QBCreateMethod


from decimal import Decimal 
from pprint import pprint 
from typing import List
from qbodbc.utils import Ref
from qbodbc.objects.base_object import BaseObject 

class BillItemLine(BaseObject): 
    """
    Add a line item to a bill. Minimum required fields. Other items: 
        - class: ItemLineClassRefFullName
        - customer: ItemLineCustomerRefFullName
        - customer:job: ItemLineCustomerRefFullName
        - cogs account: ItemLineOverrideItemAccountRefFullName  
    """
    __table_name__ = 'BillItemLine'

    def __init__(self, Item: str, Desc: str, Quantity: Decimal, Amount: Decimal, **kwargs): 
        self.ItemLineItemRefFullName = Item
        self.ItemLineDesc = Desc
        self.ItemLineQuantity = Quantity
        self.ItemLineCost = Amount
        self.FQSaveToCache = 1 # Default to save 

        self.ItemLineItemRefListId = kwargs.pop('ItemLineItemRefListId', None)
        self.ItemLineCustomerRefFullName = kwargs.pop('ItemLineCustomerRefFullName', None)
        self.ItemLineOverrideItemAccountRefFullName = kwargs.pop('ItemLineOverrideItemAccountRefFullName', None)
        self.ItemLineClassRefFullName = kwargs.pop('ItemLineClassRefFullName', None)

class BillExpenseLine(BaseObject):
    __table_name__ = 'BillExpenseLine'

    def __init__(self, account: str, memo: str, amount: Decimal, **kwargs): 
        self.ExpenseLineAccountRefFullName = account
        self.ExpenseLineMemo = memo
        self.ExpenseLineAmount = amount
 

class Bill(BaseObject):
    __table_name__ = 'Bill'

    def __init__(self, **kwargs): 
        """Bill with minimum field requirements.""" 
        self.VendorRefFullName = kwargs.pop('Vendor', None)
        self.APAccountRefFullName = kwargs.pop('APAccount', None)
        self.TxnDate = kwargs.pop('TxnDate', None)
        self.RefNumber = kwargs.pop('RefNumber', None)
        self.Memo = kwargs.pop('Memo', None)
        # [setattr(self, k, v) for k,v in kwargs.items()]
        
        # List of dictionaries 
        self._line_item = [] 
        self._line_expense = [] 

    def add_item(self, item):
        """
        Takes a list of dictionaries or a single dictionary
        """
        if isinstance(item, list): 
            [self._line_item.append(i) for i in item]
        else: 
            self._line_item.append(item)

    def add_expense(self, expense): 
        """Add an expense item to a bill."""
        if isinstance(expense, list): 
            [self._line_expense.append(i) for i in expense]
        else: 
            self._line_expense.append(expense)

    def save(self, qb=''):
        """        
        If an item, and expense are required, the bill must exist before 
        the new insert can be preformed. 

        Overide the default method since it needs to work on an iterable. 

        Pop an item in the list and assign the current bill args to it. 
        """
        try: 
            while self._line_item: 
                print('***', len(self._line_item), '***')

                head = self._line_item.pop(0)
                head.update_obj(self)

                if len(self._line_item) == 0:    
                    # Update FQSaveToCache == 0 
                    head.FQSaveToCache = 0
                    qb.cursor.execute(head.create(), head.to_values())
                else: 
                    qb.cursor.execute(head.create(), head.to_values())

            return {
                'Vendor': self.VendorRefFullName, 
                'RefNum': self.RefNumber,
                # Last insert on a parent table will still refrence an insert on the child table
                "id": qb.last_insert(self.__table_name__)
                }

        except Exception as e:
            return {
                "id": e.args[-1],
                "key": self.RefNumber
            }
       
       
    @classmethod
    def get_args(cls, con): 
        try: 
            bill_args = con.query("sp_columns Bill")
            args = bill_args[bill_args["IS_NULLABLE"] == "YES"][["COLUMNNAME", "TYPENAME"]]
            pprint(args)
            cls.args = args
        except Exception as e: 
            print("Error: ", e)

    @classmethod 
    def get_table(cls, con):
        """Cache results if a read is done."""
        cls._table = con.query("SELECT * FROM Bill")


    def __repr__(self):
        return f"""
            Bill: {self.VendorRefFullName}: {self.RefNumber}
            Line Items: {len(self._line_item)} Expense Items: {len(self._line_expense)}
            """



# Inherit from BaseCredit, BaseDebit
# Should handle 1 or many lines 
# Check for balance 

class Journal: 
    __table__ = 'Journal'

    def __init__(self): 
        self.credit_line = []
        self.debit_line = []

    def add_credit_line(self, **kwargs): 
        self.credit_line.append(
            {
                **{k:v for k,v in kwargs.items() if v is not None},
                'FQSaveToCache': '1'
                }
            )

    def add_debit_line(self, **kwargs): 
        self.debit_line.append({**kwargs, 'FQSaveToCache': '0'})
        
    def create(self, qb, **kwargs): 
        """
        credit = {
            'TxnDate': datetime.date(2021,1, 1),
            'JournalCreditLineAccountRefListID': None,
            'JournalCreditLineAccountRefFullName': 'Accounts Payable',
            'JournalCreditLineAmount': Decimal('10.99'),
            'JournalCreditLineMemo': 'MyCreditMemo',
            'JournalCreditLineEntityRefListID': None,
            'JournalCreditLineEntityRefFullName': 'Locksmith',
            'JournalCreditLineEntityClassRefListID': None,
            'JournalCreditLineEntityClassRefFullName': None,
        }
        """
        for credit in self.credit_line:
            centry = (
                f"insert into JournalEntryCreditLine ({', '.join(credit.keys())})"
                f" values ({','.join('?' * len(credit))})"
            ) 
            qb.execute(centry, *list(credit.values()))
            print(centry)

        for debit in self.debit_line: 
            dentry = (
                f"insert into JournalEntryDebitLine ({', '.join(debit.keys())})"
                f" values ({','.join('?' * len(debit))})"
            ) 
            qb.execute(dentry, *list(debit.values()))
            print(dentry)
        
        return qb.execute('SP_LASTINSERTID JournalEntryDebitLine')





class Address: 
    def __init__(self, **kwargs): 
        self.BillAddressAddr1 = kwargs.pop('BillAddressAddr1', None)
        self.BillAddressAddr2 = kwargs.pop('BillAddressAddr2', None)
        self.BillAddressAddr3 = kwargs.pop('BillAddressAddr3', None)
        self.BillAddressAddr4 = kwargs.pop('BillAddressAddr4', None)
        self.BillAddressAddr5 = kwargs.pop('BillAddressAddr5', None)
        self.BillAddressCity = kwargs.pop('BillAddressCity', None)
        self.BillAddressState = kwargs.pop('BillAddressState', None)
        self.BillAddressProvince = kwargs.pop('BillAddressProvince', None)
        self.BillAddressCounty = kwargs.pop('BillAddressCounty', None)
        self.BillAddressPostalCode = kwargs.pop('BillAddressPostalCode', None)
        self.BillAddressCountry = kwargs.pop('BillAddressCountry', None)
     
        self.ShipAddressAddr1 = kwargs.pop('ShipAddressAddr1', None)
        self.ShipAddressAddr2 = kwargs.pop('ShipAddressAddr2', None)
        self.ShipAddressAddr3 = kwargs.pop('ShipAddressAddr3', None)
        self.ShipAddressAddr4 = kwargs.pop('ShipAddressAddr4', None)
        self.ShipAddressAddr5 = kwargs.pop('ShipAddressAddr5', None)
        self.ShipAddressCity = kwargs.pop('ShipAddressCity', None)
        self.ShipAddressState = kwargs.pop('ShipAddressState', None)
        self.ShipAddressProvince = kwargs.pop('ShipAddressProvince', None)
        self.ShipAddressCounty = kwargs.pop('ShipAddressCounty', None)
        self.ShipAddressPostalCode = kwargs.pop('ShipAddressPostalCode', None)
        self.ShipAddressCountry = kwargs.pop('ShipAddressCountry', None)
       
        self.Phone = kwargs.pop('Phone', None)
        self.AltPhone = kwargs.pop('AltPhone', None)
        self.Fax = kwargs.pop('Fax', None)
        self.Email = kwargs.pop('Email', None)
        self.Cc = kwargs.pop('Cc', None)
        self.Contact = kwargs.pop('Contact', None)
      


from qbodbc.objects.address import Address
from qbodbc.objects.base_object import BaseObject
from qbodbc.utils import Ref

class Customer(BaseObject, Address): 

    __table_name__ = 'customer'
    __table_id__ = 'Name'
    
    def __init__(self, **kwargs):
        """
        Customer Model:
            Initialize a customer with the minimum requirements.
        """
        super(Customer, self).__init__(**kwargs)
        self.Name=kwargs.pop('Name', None)
        self.AccountNumber=kwargs.pop('AccountNumber', None)
        self.CompanyName=kwargs.pop('CompanyName', None)
        self.FirstName=kwargs.pop('FirstName', None)
        self.MiddleName=kwargs.pop('MiddleName', None)
        self.LastName=kwargs.pop('LastName', None)
        self.JobTitle=kwargs.pop('JobTitle', None) 
        self.CustomerTypeRefFullName=kwargs.pop('CustomerTypeRefFullName', None)
        self.TermsRefFullName=kwargs.pop('TermsRefFullName', None)
        self.SalesRepRefFullName=kwargs.pop('SalesRepRefFullName', None)
        self.ResaleNumber=kwargs.pop('ResaleNumber', None)
        self.BusinessNumber=kwargs.pop('BusinessNumber', None)
        self.CreditLimit=kwargs.pop('CreditLimit', None)
        self.JobStatus=kwargs.pop('JobStatus', None)
        self.JobStartDate=kwargs.pop('JobStartDate', None)
        self.JobProjectedEndDate=kwargs.pop('JobProjectedEndDate', None)
        self.JobEndDate=kwargs.pop('JobEndDate', None)
        self.JobDesc=kwargs.pop('JobDesc', None)
        self.JobTypeRefListID=kwargs.pop('JobTypeRefListID', None)
        self.JobTypeRefFullName=kwargs.pop('JobTypeRefFullName', None)
        self.Notes=kwargs.pop('Notes', None)
        self.ExternalGUID=kwargs.pop('ExternalGUID', None)
        #  'FullName',
        #  'IsActive',
        #  'Salutation',
        #  'CustomerTypeRefListID',
        #  'TermsRefListID',
        #  'SalesRepRefListID',
        #  'Balance',
        #  'TotalBalance',
        #  'OpenBalance',
        #  'OpenBalanceDate',
        #  'SalesTaxCodeRefListID',
        #  'SalesTaxCodeRefFullName',
        #  'TaxCodeRefListID',
        #  'TaxCodeRefFullName',
        #  'ItemSalesTaxRefListID',
        #  'ItemSalesTaxRefFullName',
        #  'SalesTaxCountry',
        #  'PreferredPaymentMethodRefListID',
        #  'PreferredPaymentMethodRefFullName',
        #  'CreditCardInfoCreditCardNumber',
        #  'CreditCardInfoExpirationMonth',
        #  'CreditCardInfoExpirationYear',
        #  'CreditCardInfoNameOnCard',
        #  'CreditCardInfoCreditCardAddress',
        #  'CreditCardInfoCreditCardPostalCode',
        #  'PreferredDeliveryMethod',
        #  'IsUsingCustomerTaxCode',
        #  'PriceLevelRefListID',
        #  'PriceLevelRefFullName',
        #  'TaxRegistrationNumber',
        #  'CurrencyRefListID',
        #  'CurrencyRefFullName'
        
    @property 
    def to_ref(self):
        ref = Ref()
        ref.Name = 'Name'
        ref.Value = self.Name
        return ref

    def job(self, job):
        """
        Create a customer or query a customer to use as a ref with to_ref()
        """
        ...

    def __str__(self): 
        return self.Name


import datetime 
from decimal import Decimal 
import random 
import secrets

# Don't need to format date!!! Will accept date time attribute! 
# def format_date(date): 
#     if type(date) == datetime.date:
#         return "d{" + f"'{date.strftime('%Y-%m-%d')}'" + "}"
#     else:
#         return "d{" + f"'{date}'" + "}"


# def update_date(vals):
#     return [format_date(i).encode("utf-8") if type(i) == datetime.date else '?' for i in vals]

def format_decimal(vals): 
    """Encoding needs to be supplied for floats and integers."""
    rsp = []
    for i in vals:
        try: 
            rsp.append(Decimal(i))
        except: 
            rsp.append("?")
    return rsp 


def as_decimal(x):
    return Decimal(x).__round__(2)

def as_date(y, m, d):
    return datetime.date(y, m, d)


def name_generator(): 
    first_name = [
        'bob', 'marty', 'marsha', 'rasham', 'greg', 'gregory',
        'steve', 'mary', 'bill', 'suf', 'ashley', 'bently',
        'vicky', 'neria', 'moo', 'cow', 'dog', 'hary', 'fenti'
    ]
    last_name = [
        'meek', 'bills', 'type', 'popcorn', 'wordsss', 'blake'
        'blane', 'ash', 'blaire', 'world', 'fair', 'yarn', 'plow', 
        'rank', 'rafiki', 'reynolds', 'bear'
        ]

    return f"{random.choice(first_name)} {random.choice(last_name)}"


def account_generator(): 
    return str(secrets.randbits(29))


class Ref: 
    def __init__(self): 
        self.Name = ""
        self.Value = ""
        



# list mixin 
# detail mixin | filter 
# create, update, destroy mixin
# from qbodbc.client import Client

def to_dict(obj, classkey=None):
    """
    Recursively converts Python object into a dictionary
    Stolen from: https://github.com/ej2/python-quickbooks/blob/c82935bd07d9901d0e8d10765a7fcb06ad2ef53d/quickbooks/mixins.py#L89
    """
    if isinstance(obj, dict):
        data = {}
        for (k, v) in obj.items():
            data[k] = to_dict(v, classkey)
        return data
    elif hasattr(obj, "_ast"):
        return to_dict(obj._ast())
    elif hasattr(obj, "__iter__") and not isinstance(obj, str):
        return [to_dict(v, classkey) for v in obj]
    elif hasattr(obj, "__dict__"):
        data = dict([(key, to_dict(value, classkey))
                    for key, value in obj.__dict__.items()
                    if not callable(value) and not key.startswith('_')])

        if classkey is not None and hasattr(obj, "__class__"):
            data[classkey] = obj.__class__.__name__
        return data
    else:
        return obj

class ToDictMixin:
    def to_dict(self):
        return to_dict(self)


class BaseObject(ToDictMixin): 
           
    def to_names(self): 
        """
        Turns object attribute names into comma seperated string to be used 
        in insert statements.
        """
        return ', '.join([key for key, value in self.to_dict().items() if value is not None])
    
    def to_values(self): 
        """
        Turns object attribute values into list of values to be used in 
        insert statements.
        """
        return [value for key, value in self.to_dict().items() if value is not None]
    
    def to_params(self): 
        """
        Creates a comma seperated string of question marks to be used 
        in paramterized queries. The number of question marks returns should 
        match the length of to_values().
        """
        return ', '.join('?' * len([key for key, value in self.to_dict().items() if value is not None]))

    def update_obj(self, obj):
        """
        Merge a child object with its parent or vica versa. 
        """
        for k, v in obj.to_dict().items():
            if v:
                if v is not None:
                    setattr(self, k, v) 

    def process_response(): 
        # Something to handle errors in a uniform manner
        pass 

    def create(self):
        q = (
            f"insert into {self.__table_name__} " 
                f"({self.to_names()}) values " 
            f"({self.to_params()})"
        )

        return q

    # def update(self): 
    #     """
    #         update customer set 
    #             accountnumber = ?, customfield_xxx = ? 
    #         where listid = ? 
    #     """
    #     pass 


    def delete(self): 
        pass 


    def save(self, qb):
        """
        Create or update
        Args: 
            qb: session object, instance of QuickBooks().connect().
        """
        try:
            qb.cursor.execute(self.create(), self.to_values())
                
            return {
                "id": qb.last_insert(self.__table_name__),
                "key": self.to_ref.Value
            }

        except Exception as e:
            return {
                "id": e.args[-1],
                "key": self.to_ref.Value
            }
    

    # Should all be passed as mixins to the base object. 
    def get(self, qb): 
        """
        Returns one or None. To be used in updates. 
        """
        pass 

    def all(self, qb, params): 
        """
        Gets the entire table. Takes optional param of column names to return. 
        """
        pass 
    
    def filter(self, qb, params): 
        """
        Return a subset of customers that meet certian conditions. 
        """
        pass 

